#
# Copyright (C) 2016-2017 University of Oxford
#
# This file is part of msprime.
#
# msprime is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# msprime is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with msprime.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Common provenance methods used to determine the state and versions
of various dependencies and the OS.
"""
import base64
import importlib
import json
import logging
import marshal
import types

import numpy
import tskit

from . import ancestry
from msprime import _msprime

__version__ = "undefined"
try:
    from . import _version

    __version__ = _version.version
except ImportError:
    pass
logger = logging.getLogger(__name__)


class CURRENT_TREE_SEQUENCE:
    pass


def get_provenance_dict(parameters=None):
    """
    Returns a dictionary encoding an execution of msprime conforming to the
    tskit provenance schema.
    """
    document = {
        "schema_version": "1.0.0",
        "software": {"name": "msprime", "version": __version__},
        "parameters": parameters,
        "environment": get_environment(),
    }
    return document


def _get_environment():
    gsl_version = ".".join(map(str, _msprime.get_gsl_version()))
    libraries = {"gsl": {"version": gsl_version}}
    return tskit.provenance.get_environment(extra_libs=libraries)


_environment = None


def get_environment():
    """
    Returns a dictionary describing the environment in which tskit
    is currently running.
    """
    # Everything here is fixed so we cache it
    global _environment
    if _environment is None:
        _environment = _get_environment()
    return _environment


class ProvenanceEncoderDecoder(json.JSONEncoder):
    """
    Extension of the `json` encoder that serialises arbitrary python class objects
    by calling an `asdict` method on the object that should provide the arguments
    needed to exactly recreate it. Special cases numpy arrays, classes and functions
    """

    def default(self, obj):
        # NOTE: this is more complicated than it needs to be now because the
        # SimulationModelChange is simpler than it was when this was developed.
        # We no longer have to deal with function arguments. The version we
        # worked against here was never released.

        # TODO it's a bit ugly and brittle that we have to have these
        # special cases, but this needed so that we can recursively
        # process the SimulationModelChange below.
        if obj is None or isinstance(obj, str):
            return obj
        if isinstance(obj, types.FunctionType):
            # Some provenance-relevant classes such as SimulationModelChange take
            # functional arguments. Note that we have a tight definition of
            # what we allow as a function here. We don't include builtins or
            # callable classes, we also block functions that refer to globals
            # or the outer scope for both complexity and security reasons.
            # We can deserialise these later with
            # FunctionType(fcode, fglobals, fname, fdefaults, fclosure)
            if len(obj.__code__.co_names) > 0:
                error = (
                    f"Configuration function {obj.__code__.co_name} refers to"
                    f" global variables {obj.__code__.co_names} which are not"
                    f" currently serialised"
                )
                logger.warning(error)
                return {"__error__": error}
            if obj.__closure__ is not None and len(obj.__closure__) > 0:
                error = (
                    f"Configuration function {obj.__code__.co_name} refers to outer"
                    f" scope variables which are not currently serialised"
                )
                logger.warning(error)
                return {"__error__": error}
            return {
                "__function__": base64.b64encode(marshal.dumps(obj.__code__)).decode(
                    "utf-8"
                ),
                "defaults": obj.__defaults__,
            }

        if isinstance(obj, (tskit.TreeSequence, tskit.TableCollection)):
            # If a tree sequence is an argument to a function, then that tree sequence
            # can be recreated from its provenance, which will be the provenance record
            # before the one produced by the function being called. Hence we do not
            # store the tree sequence in provenance.
            return {"__constant__": "__current_ts__"}

        elif isinstance(obj, numpy.ndarray):
            # The most failsafe way would be be to `base64.b64encode(obj.tostring())`
            # but as we expect only simple arrays we make this trade-off so that
            # the array is human readable
            return {"__ndarray__": obj.tolist(), "dtype": obj.dtype.str}

        elif isinstance(obj, numpy.number):
            # Some numbers come through as numpy types
            return {
                "__npgeneric__": str(obj),
                "dtype": obj.dtype.str,
            }

        try:
            ret = obj.asdict()
            cls = obj.__class__
            ret["__class__"] = f"{cls.__module__}.{cls.__name__}"
            if isinstance(obj, ancestry.SimulationModelChange):
                # We have to special-case the SimulationModelChange because it
                # can contain an embedded msprime object. To complicate things
                # a bit more, the model can be either None, a string or an
                # instance. The simplest way around this seems to be to
                # recursively parse the process the model and to guard against
                # None as input
                ret["model"] = self.default(obj.model)
            return ret
        except AttributeError:
            raise TypeError(
                f"Object of type {obj.__class__.__name__} "
                f"is not JSON serialisable. Please provide an `asdict` method "
                f"that returns the object's constructor arguments."
            )

    @staticmethod
    def decode(s):
        def hook(obj):
            if "__function__" in obj:
                return types.FunctionType(
                    marshal.loads(base64.b64decode(obj["__function__"])),
                    {},
                    "func",
                    obj["defaults"],
                    (),
                )
            elif "__constant__" in obj:
                return {"__current_ts__": CURRENT_TREE_SEQUENCE}[obj["__constant__"]]
            elif "__ndarray__" in obj:
                return numpy.asarray(obj["__ndarray__"], dtype=obj["dtype"])
            elif "__npgeneric__" in obj:
                return numpy.array([obj["__npgeneric__"]]).astype(obj["dtype"])[0]
            elif "__class__" in obj:
                module, cls = obj["__class__"].rsplit(".", 1)
                module = importlib.import_module(module)
                del obj["__class__"]
                return getattr(module, cls)(**obj)
            return obj

        return json.JSONDecoder(object_hook=hook).decode(s)


def parse_provenance(provenance, current_ts):
    """
    Convert the specified provenance to tuple of the command used and a dict suitable
    for calling the command again e.g:
    `msprime.simulate(**parse_provenance(provenance, None)[1])`
    """
    ret = ProvenanceEncoderDecoder.decode(provenance.record)
    if ret["software"]["name"] != "msprime":
        raise ValueError(
            f"Only msprime provenances can be parsed,"
            f' found {ret["software"]["name"]}'
        )
    parameters = ret["parameters"]
    command = parameters.pop("command")
    updated_parameters = {}
    for key, value in parameters.items():
        if value == CURRENT_TREE_SEQUENCE:
            value = current_ts
        updated_parameters[key] = value
    return command, updated_parameters


def _human_readable_size(size, decimal_places=2):
    chosen_unit = "TB"
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024.0:
            chosen_unit = unit
            break
        size /= 1024.0
    return f"{size:.{decimal_places}f}{chosen_unit}"


def json_encode_provenance(provenance_dict, num_replicates=1):
    """
    Return a JSON representation of the provenance
    """
    prov = ProvenanceEncoderDecoder().encode(provenance_dict)
    if len(prov) > 2_097_152:
        logger.warning(
            f"The provenance information for the resulting tree sequence is"
            f" {_human_readable_size(len(prov))}."
            f" This is nothing to worry about as provenance is a good thing"
            f" to have, but if you want to save this memory/storage space"
            f" you can disable provenance recording by setting"
            f" record_provenance=False"
        )
    return prov



#
# Copyright (C) 2020-2021 University of Oxford
#
# This file is part of msprime.
#
# msprime is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# msprime is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with msprime.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Utilities for working with intervals and interval maps.
"""
from __future__ import annotations

import collections.abc
import itertools
import numbers
import warnings

import numpy as np

from msprime import core


class RateMap(collections.abc.Mapping):
    """
    A class mapping a non-negative rate value to a set of non-overlapping intervals
    along the genome. Intervals for which the rate is unknown (i.e., missing data)
    are encoded by NaN values in the ``rate`` array.

    :param list position: A list of :math:`n+1` positions, starting at 0, and ending
        in the sequence length over which the RateMap will apply.
    :param list rate: A list of :math:`n` positive rates that apply between each
        position. Intervals with missing data are encoded by NaN values.
    """

    # The args are marked keyword only to give us some flexibility in how we
    # create class this in the future.
    def __init__(
        self,
        *,
        position,
        rate,
    ):
        # Making the arrays read-only guarantees rate and cumulative mass stay in sync
        # We prevent the arrays themselves being overwritten by making self.position,
        # etc properties.

        # TODO we always coerce the position type to float here, but we may not
        # want to do this. int32 is a perfectly good choice a lot of the time.
        self._position = np.array(position, dtype=float)
        self._position.flags.writeable = False
        self._rate = np.array(rate, dtype=float)
        self._rate.flags.writeable = False
        size = len(self._position)
        if size < 2:
            raise ValueError("Must have at least two positions")
        if len(self._rate) != size - 1:
            raise ValueError(
                "Rate array must have one less entry than the position array"
            )
        if self._position[0] != 0:
            raise ValueError("First position must be zero")

        span = self.span
        if np.any(span <= 0):
            bad_pos = np.where(span <= 0)[0] + 1
            raise ValueError(
                f"Position values not strictly increasing at indexes {bad_pos}"
            )
        if np.any(self._rate < 0):
            bad_rates = np.where(self._rate < 0)[0]
            raise ValueError(f"Rate values negative at indexes {bad_rates}")
        self._missing = np.isnan(self.rate)
        self._num_missing_intervals = np.sum(self._missing)
        if self._num_missing_intervals == len(self.rate):
            raise ValueError("All intervals are missing data")
        # We don't expose the cumulative mass array as a part of the array
        # API is it's not quite as obvious how it lines up for each interval.
        # It's really the sum of the mass up to but not including the current
        # interval, which is a bit confusing. Probably best to just leave
        # it as a function, so that people can sample at regular positions
        # along the genome anyway, emphasising that it's a continuous function,
        # not a step function like the other interval attributes.
        self._cumulative_mass = np.insert(np.nancumsum(self.mass), 0, 0)
        assert self._cumulative_mass[0] == 0
        self._cumulative_mass.flags.writeable = False

    @property
    def left(self):
        """
        The left position of each interval (inclusive).
        """
        return self._position[:-1]

    @property
    def right(self):
        """
        The right position of each interval (exclusive).
        """
        return self._position[1:]

    @property
    def mid(self):
        """
        Returns the midpoint of each interval.
        """
        mid = self.left + self.span / 2
        mid.flags.writeable = False
        return mid

    @property
    def span(self):
        """
        Returns the span (i.e., ``right - left``) of each of the intervals.
        """
        span = self.right - self.left
        span.flags.writeable = False
        return span

    @property
    def position(self):
        """
        The breakpoint positions between intervals. This is equal to the
        :attr:`~.RateMap.left` array with the :attr:`sequence_length`
        appended.
        """
        return self._position

    @property
    def rate(self):
        """
        The rate associated with each interval. Missing data is encoded
        by NaN values.
        """
        return self._rate

    @property
    def mass(self):
        r"""
        The "mass" of each interval, defined as the :attr:`~.RateMap.rate`
        :math:`\times` :attr:`~.RateMap.span`. This is NaN for intervals
        containing missing data.
        """
        return self._rate * self.span

    @property
    def missing(self):
        """
        A boolean array encoding whether each interval contains missing data.
        Equivalent to ``np.isnan(rate_map.rate)``
        """
        return self._missing

    @property
    def non_missing(self):
        """
        A boolean array encoding whether each interval contains non-missing data.
        Equivalent to ``np.logical_not(np.isnan(rate_map.rate))``
        """
        return ~self._missing

    #
    # Interval counts
    #

    @property
    def num_intervals(self) -> int:
        """
        The total number of intervals in this map. Equal to
        :attr:`~.RateMap.num_missing_intervals` +
        :attr:`~.RateMap.num_non_missing_intervals`.
        """
        return len(self._rate)

    @property
    def num_missing_intervals(self) -> int:
        """
        Returns the number of missing intervals, i.e., those in which the
        :attr:`~.RateMap.rate` value is a NaN.
        """
        return self._num_missing_intervals

    @property
    def num_non_missing_intervals(self) -> int:
        """
        The number of non missing intervals, i.e., those in which the
        :attr:`~.RateMap.rate` value is not a NaN.
        """
        return self.num_intervals - self.num_missing_intervals

    @property
    def sequence_length(self):
        """
        The sequence length covered by this map
        """
        return self.position[-1]

    @property
    def total_mass(self):
        """
        The cumulative total mass over the entire map.
        """
        return self._cumulative_mass[-1]

    @property
    def mean_rate(self):
        """
        The mean rate over this map weighted by the span covered by each rate.
        Unknown intervals are excluded.
        """
        total_span = np.sum(self.span[self.non_missing])
        return self.total_mass / total_span

    def get_rate(self, x):
        """
        Return the rate at the specified list of positions.

        .. note:: This function will return a NaN value for any positions
            that contain missing data.

        :param numpy.ndarray x: The positions for which to return values.
        :return: An array of rates, the same length as ``x``.
        :rtype: numpy.ndarray
        """
        loc = np.searchsorted(self.position, x, side="right") - 1
        if np.any(loc < 0) or np.any(loc >= len(self.rate)):
            raise ValueError("position out of bounds")
        return self.rate[loc]

    def get_cumulative_mass(self, x):
        """
        Return the cumulative mass of the map up to (but not including) a
        given point for a list of positions along the map. This is equal to
        the integral of the rate from 0 to the point.

        :param numpy.ndarray x: The positions for which to return values.

        :return: An array of cumulative mass values, the same length as ``x``
        :rtype: numpy.ndarray
        """
        x = np.array(x)
        if np.any(x < 0) or np.any(x > self.sequence_length):
            raise ValueError(f"Cannot have positions < 0 or > {self.sequence_length}")
        return np.interp(x, self.position, self._cumulative_mass)

    def find_index(self, x: float) -> int:
        """
        Returns the index of the interval that the specified position falls within,
        such that ``rate_map.left[index] <= x < self.rate_map.right[index]``.

        :param float x: The position to search.
        :return: The index of the interval containing this point.
        :rtype: int
        :raises: KeyError if the position is not contained in any of the intervals.
        """
        if x < 0 or x >= self.sequence_length:
            raise KeyError(f"Position {x} out of bounds")
        index = np.searchsorted(self.position, x, side="left")
        if x < self.position[index]:
            index -= 1
        assert self.left[index] <= x < self.right[index]
        return index

    def missing_intervals(self):
        """
        Returns the left and right coordinates of the intervals containing
        missing data in this map as a 2D numpy array
        with shape (:attr:`~.RateMap.num_missing_intervals`, 2). Each row
        of this returned array is therefore a ``left``, ``right`` tuple
        corresponding to the coordinates of the missing intervals.

        :return: A numpy array of the coordinates of intervals containing
            missing data.
        :rtype: numpy.ndarray
        """
        out = np.empty((self.num_missing_intervals, 2))
        out[:, 0] = self.left[self.missing]
        out[:, 1] = self.right[self.missing]
        return out

    def asdict(self):
        return {"position": self.position, "rate": self.rate}

    #
    # Dunder methods. We implement the Mapping protocol via __iter__, __len__
    # and __getitem__. We have some extra semantics for __getitem__, providing
    # slice notation.
    #

    def __iter__(self):
        # The clinching argument for using mid here is that if we used
        # left instead we would have
        #   RateMap([0, 1], [0.1]) == RateMap([0, 100], [0.1])
        # by the inherited definition of equality since the dictionary items
        # would be equal.
        # Similarly, we only return the midpoints of known intervals
        # because NaN values are not equal, and we would need to do
        # something to work around this. It seems reasonable that
        # this high-level operation returns the *known* values only
        # anyway.
        yield from self.mid[self.non_missing]

    def __len__(self):
        return np.sum(self.non_missing)

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step is not None:
                raise TypeError("Only interval slicing is supported")
            return self.slice(key.start, key.stop)
        if isinstance(key, numbers.Number):
            index = self.find_index(key)
            if np.isnan(self.rate[index]):
                # To be consistent with the __iter__ definition above we
                # don't consider these missing positions to be "in" the map.
                raise KeyError(f"Position {key} is within a missing interval")
            return self.rate[index]
        # TODO we could implement numpy array indexing here and call
        # to get_rate. Note we'd need to take care that we return a keyerror
        # if the returned array contains any nans though.
        raise KeyError("Key {key} not in map")

    def _display_table(self):
        def format_row(left, right, mid, span, rate):
            return [
                f"{left:.10g}",
                f"{right:.10g}",
                f"{mid:.10g}",
                f"{span:.10g}",
                f"{rate:.2g}",
            ]

        def format_slice(start, end):
            return list(
                itertools.starmap(
                    format_row,
                    zip(
                        self.left[start:end],
                        self.right[start:end],
                        self.mid[start:end],
                        self.span[start:end],
                        self.rate[start:end],
                    ),
                )
            )

        if self.num_intervals < 40:
            data = format_slice(0, None)
        else:
            data = format_slice(0, 10)
            data.append(["⋯"] * 5)
            data += format_slice(-10, None)

        return ["left", "right", "mid", "span", "rate"], data

    def __str__(self):
        titles, data = self._display_table()
        data = [[[item] for item in row] for row in data]
        table = core.text_table(
            caption="",
            column_titles=[[title] for title in titles],
            column_alignments="<<>>>",
            data=data,
        )
        return table

    def _repr_html_(self):
        col_titles, data = self._display_table()
        return core.html_table("", col_titles, data)

    def __repr__(self):
        return f"RateMap(position={repr(self.position)}, rate={repr(self.rate)})"

    #
    # Methods for building rate maps.
    #

    def copy(self) -> RateMap:
        """
        Returns a deep copy of this RateMap.
        """
        # We take read-only copies of the arrays in the constructor anyway, so
        # no need for copying.
        return RateMap(position=self.position, rate=self.rate)

    def slice(self, left=None, right=None, *, trim=False) -> RateMap:  # noqa: A003
        """
        Returns a subset of this rate map in the specified interval.

        :param float left: The left coordinate (inclusive) of the region to keep.
            If ``None``, defaults to 0.
        :param float right: The right coordinate (exclusive) of the region to keep.
            If ``None``, defaults to the sequence length.
        :param bool trim: If True, remove the flanking regions such that the
            sequence length of the new rate map is ``right`` - ``left``. If ``False``
            (default), do not change the coordinate system and mark the flanking
            regions as "unknown".
        :return: A new RateMap instance
        :rtype: RateMap
        """
        left = 0 if left is None else left
        right = self.sequence_length if right is None else right
        if not (0 <= left < right <= self.sequence_length):
            raise KeyError(f"Invalid slice: left={left}, right={right}")

        i = self.find_index(left)
        j = i + np.searchsorted(self.position[i:], right, side="right")
        if right > self.position[j - 1]:
            j += 1

        position = self.position[i:j].copy()
        rate = self.rate[i : j - 1].copy()
        position[0] = left
        position[-1] = right

        if trim:
            # Return trimmed map with changed coords
            return RateMap(position=position - left, rate=rate)

        # Need to check regions before & after sliced region are filled out:
        if left != 0:
            if np.isnan(rate[0]):
                position[0] = 0  # Extend
            else:
                rate = np.insert(rate, 0, np.nan)  # Prepend
                position = np.insert(position, 0, 0)
        if right != self.position[-1]:
            if np.isnan(rate[-1]):
                position[-1] = self.sequence_length  # Extend
            else:
                rate = np.append(rate, np.nan)  # Append
                position = np.append(position, self.position[-1])
        return RateMap(position=position, rate=rate)

    @staticmethod
    def uniform(sequence_length, rate) -> RateMap:
        """
        Create a uniform rate map
        """
        return RateMap(position=[0, sequence_length], rate=[rate])

    @staticmethod
    def read_hapmap(
        fileobj,
        sequence_length=None,
        *,
        has_header=True,
        position_col=None,
        rate_col=None,
        map_col=None,
    ):
        # Black barfs with an INTERNAL_ERROR trying to reformat this docstring,
        # so we explicitly disable reformatting here.
        # fmt: off
        """
        Parses the specified file in HapMap format and returns a :class:`.RateMap`.
        HapMap files must white-space-delimited, and by default are assumed to
        contain a single header line (which is ignored). Each subsequent line
        then contains a physical position (in base pairs) and either a genetic
        map position (in centiMorgans) or a recombination rate (in centiMorgans
        per megabase); the rate between the current physical position
        (inclusive) and the physical position on the next line (exclusive) is
        taken as constant. By default, the second column of the file is taken
        as the physical position and the fourth column is taken as the genetic
        position, as seen in the following sample of the format::

            Chromosome	Position(bp)  Rate(cM/Mb)  Map(cM)
            chr10       48232         0.1614       0.002664
            chr10       48486         0.1589       0.002705
            chr10       50009         0.159        0.002947
            chr10       52147         0.1574       0.003287
            ...
            chr10	133762002     3.358        181.129345
            chr10	133766368     0.000        181.144008

        In the example above, the first row has a nonzero genetic map position
        (last column, cM), implying a nonzero recombination rate before that
        position, that is assumed to extend to the start of the chromosome
        (at position 0 bp). However, if the first line has a nonzero bp position
        (second column) and a zero genetic map position (last column, cM),
        then the recombination rate before that position is *unknown*, producing
        :ref:`missing data <sec_rate_maps_missing>`.

        .. note::
            The rows are all assumed to come from the same contig, and the
            first column is currently ignored. Therefore if you have a single
            file containing several contigs or chromosomes, you must must split
            it up into multiple files, and pass each one separately to this
            function.

        :param str fileobj: Filename or file to read. This is passed directly
            to :func:`numpy.loadtxt`, so if the filename extension is .gz or .bz2,
            the file is decompressed first
        :param float sequence_length: The total length of the map. If ``None``,
            then assume it is the last physical position listed in the file.
            Otherwise it must be greater then or equal to the last physical
            position in the file, and the region between the last physical position
            and the sequence_length is padded with a rate of zero.
        :param bool has_header: If True (default), assume the file has a header row
            and ignore the first line of the file.
        :param int position_col: The zero-based index of the column in the file
            specifying the physical position in base pairs. If ``None`` (default)
            assume an index of 1 (i.e. the second column).
        :param int rate_col: The zero-based index of the column in the file
            specifying the rate in cM/Mb. If ``None`` (default) do not use the rate
            column, but calculate rates using the genetic map positions, as
            specified in ``map_col``. If the rate column is used, the
            interval from 0 to first physical position in the file is marked as
            unknown, and the last value in the rate column must be zero.
        :param int map_col: The zero-based index of the column in the file
            specifying the genetic map position in centiMorgans. If ``None``
            (default), assume an index of 3 (i.e. the fourth column). If the first
            genetic position is 0 the interval from position 0 to the first
            physical position in the file is marked as unknown. Otherwise, act
            as if an additional row, specifying physical position 0 and genetic
            position 0, exists at the start of the file.
        :return: A RateMap object.
        :rtype: RateMap
        """
        # fmt: on
        column_defs = {}  # column definitions passed to np.loadtxt
        if rate_col is None and map_col is None:
            # Default to map_col
            map_col = 3
        elif rate_col is not None and map_col is not None:
            raise ValueError("Cannot specify both rate_col and map_col")
        if map_col is not None:
            column_defs[map_col] = ("map", float)
        else:
            column_defs[rate_col] = ("rate", float)
        position_col = 1 if position_col is None else position_col
        if position_col in column_defs:
            raise ValueError(
                "Cannot specify the same columns for position_col and "
                "rate_col or map_col"
            )
        column_defs[position_col] = ("pos", int)

        column_names = [c[0] for c in column_defs.values()]
        column_data = np.loadtxt(
            fileobj,
            skiprows=1 if has_header else 0,
            dtype=list(column_defs.values()),
            usecols=list(column_defs.keys()),
            unpack=True,
        )
        data = dict(zip(column_names, column_data))

        if "map" not in data:
            assert "rate" in data
            if data["rate"][-1] != 0:
                raise ValueError("The last entry in the 'rate' column must be zero")
            pos_Mb = data["pos"] / 1e6
            map_pos = np.cumsum(data["rate"][:-1] * np.diff(pos_Mb))
            data["map"] = np.insert(map_pos, 0, 0) / 100
        else:
            data["map"] /= 100  # Convert centiMorgans to Morgans
        if len(data["map"]) == 0:
            raise ValueError("Empty hapmap file")

        # TO DO: read in chrom name from col 0 and poss set as .name
        # attribute on the RateMap

        physical_positions = data["pos"]
        genetic_positions = data["map"]
        start = physical_positions[0]
        end = physical_positions[-1]

        if genetic_positions[0] > 0 and start == 0:
            raise ValueError(
                "The map distance at the start of the chromosome must be zero"
            )
        if start > 0:
            physical_positions = np.insert(physical_positions, 0, 0)
            if genetic_positions[0] > 0:
                # Exception for a map that starts > 0cM: include the start rate
                # in the mean
                start = 0
            genetic_positions = np.insert(genetic_positions, 0, 0)

        if sequence_length is not None:
            if sequence_length < end:
                raise ValueError(
                    "The sequence_length cannot be less than the last physical position "
                    f" ({physical_positions[-1]})"
                )
            if sequence_length > end:
                physical_positions = np.append(physical_positions, sequence_length)
                genetic_positions = np.append(genetic_positions, genetic_positions[-1])

        assert genetic_positions[0] == 0
        rate = np.diff(genetic_positions) / np.diff(physical_positions)
        if start != 0:
            rate[0] = np.nan
        if end != physical_positions[-1]:
            rate[-1] = np.nan
        return RateMap(position=physical_positions, rate=rate)


class RecombinationMap:
    """
    A RecombinationMap represents the changing rates of recombination
    along a chromosome. This is defined via two lists of numbers:
    ``positions`` and ``rates``, which must be of the same length.
    Given an index j in these lists, the rate of recombination
    per base per generation is ``rates[j]`` over the interval
    ``positions[j]`` to ``positions[j + 1]``. Consequently, the first
    position must be zero, and by convention the last rate value
    is also required to be zero (although it is not used).

    .. important::
        This class is deprecated (but supported indefinitely);
        please use the :class:`.RateMap` class in new code.
        In particular, note that when specifying ``rates`` in the
        the :class:`.RateMap` class we now require an array
        of length :math:`n - 1` (this class requires an array
        of length :math:`n` in which the last entry is zero).

    :param list positions: The positions (in bases) denoting the
        distinct intervals where recombination rates change. These can
        be floating point values.
    :param list rates: The list of rates corresponding to the supplied
        ``positions``. Recombination rates are specified per base,
        per generation.
    :param int num_loci: **This parameter is no longer supported.**
        Must be either None (meaning a continuous genome of the
        finest possible resolution) or be equal to ``positions[-1]``
        (meaning a discrete genome). Any other value will result in
        an error. Please see the :ref:`sec_legacy_0x_genome_discretisation`
        section for more information.
    """

    def __init__(self, positions, rates, num_loci=None, map_start=0):
        # Used as an internal flag for the 0.x simulate() function. This allows
        # us to emulate the discrete-sites behaviour of 0.x code.
        self._is_discrete = num_loci == positions[-1]
        if num_loci is not None and num_loci != positions[-1]:
            raise ValueError(
                "The RecombinationMap interface is deprecated and only "
                "partially supported. If you wish to simulate a number of "
                "discrete loci, you must set num_loci == the sequence length. "
                "If you wish to simulate recombination process on as fine "
                "a map as possible, please omit the num_loci parameter (or set "
                "to None). Otherwise, num_loci is no longer supported and "
                "the behaviour of msprime 0.x cannot be emulated. Please "
                "consider upgrading your code to the version 1.x APIs."
            )
        self.map = RateMap(position=positions, rate=rates[:-1])

    @classmethod
    def uniform_map(cls, length, rate, num_loci=None):
        """
        Returns a :class:`.RecombinationMap` instance in which the recombination
        rate is constant over a chromosome of the specified length.
        The legacy ``num_loci`` option is no longer supported and should not be used.

        :param float length: The length of the chromosome.
        :param float rate: The rate of recombination per unit of sequence length
            along this chromosome.
        :param int num_loci: This parameter is no longer supported.
        """
        return cls([0, length], [rate, 0], num_loci=num_loci)

    @classmethod
    def read_hapmap(cls, filename):
        """
        Parses the specified file in HapMap format.

        .. warning::
            This method is deprecated, use the :meth:`.RateMap.read_hapmap`
            method instead.

        :param str filename: The name of the file to be parsed. This may be
            in plain text or gzipped plain text.
        :return: A RecombinationMap object.
        """
        warnings.warn(
            "RecombinationMap.read_hapmap() is deprecated. "
            "Use RateMap.read_hapmap() instead.",
            FutureWarning,
        )
        rate_map = RateMap.read_hapmap(filename, position_col=1, rate_col=2)
        # Mark anything missing as 0 for backwards compatibility. This will
        # ensure that simulate() never trims parts of the tree sequence.
        rate = rate_map.rate.copy()
        rate[rate_map.missing] = 0
        return cls(rate_map.position, np.append(rate, 0))

    @property
    def mean_recombination_rate(self):
        """
        Return the weighted mean recombination rate
        across all windows of the entire recombination map.
        """
        return self.map.mean_rate

    def get_total_recombination_rate(self):
        """
        Returns the effective recombination rate for this genetic map.
        This is the weighted mean of the rates across all intervals.
        """
        return self.map.total_mass

    def physical_to_genetic(self, x):
        return self.map.get_cumulative_mass(x)

    def genetic_to_physical(self, genetic_x):
        if self.map.total_mass == 0:
            # If we have a zero recombination rate throughout then everything
            # except L maps to 0.
            return self.get_sequence_length() if genetic_x > 0 else 0
        if genetic_x == 0:
            return self.map.position[0]
        # TODO refactor this to this to use get_cumulative_mass() function / add the
        # corresponding high-level function to the rate map.
        index = np.searchsorted(self.map._cumulative_mass, genetic_x) - 1
        y = (
            self.map.position[index]
            + (genetic_x - self.map._cumulative_mass[index]) / self.map.rate[index]
        )
        return y

    def physical_to_discrete_genetic(self, physical_x):
        raise ValueError("Discrete genetic space is no longer supported")

    def get_per_locus_recombination_rate(self):
        raise ValueError("Genetic loci are no longer supported")

    def get_num_loci(self):
        raise ValueError("num_loci is no longer supported")

    def get_size(self):
        return len(self.map.position)

    def get_positions(self):
        return list(self.map.position)

    def get_rates(self):
        return list(self.map.rate) + [0]

    def get_sequence_length(self):
        return self.map.sequence_length

    def get_length(self):
        # Deprecated: use get_sequence_length() instead
        return self.get_sequence_length()

    def asdict(self):
        return self.map.asdict()



#
# Copyright (C) 2019-2020 University of Oxford
#
# This file is part of msprime.
#
# msprime is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# msprime is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with msprime.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Pedigree utilities.
"""
import collections
import inspect
import os

import numpy as np
import tskit


def sim_pedigree(
    *,
    population_size=None,
    sequence_length=None,
    random_seed=None,
    num_replicates=None,
    end_time=None,
):
    # Internal utility for generating pedigree data. This function is not
    # part of the public API and subject to arbitrary changes/removal
    # in the future.

    tables = tskit.TableCollection(
        sequence_length=-1 if sequence_length is None else sequence_length
    )
    # TODO we should be using the demography class here to set up the
    # population metadata
    permissive_json = tskit.MetadataSchema(
        {
            "codec": "json",
            "type": "object",
            "properties": {},
            "additionalProperties": True,
        }
    )
    tables.populations.metadata_schema = permissive_json
    # tables.individuals.metadata_schema = permissive_json
    tables.populations.add_row()

    population = np.array([tskit.NULL], dtype=np.int32)
    rng = np.random.RandomState(random_seed)

    # To make the semantics compatible with dtwf, the end_time means the
    # *end* of generation end_time
    for time in reversed(range(end_time + 1)):
        N = population_size  # This could be derived from the Demography
        # Make the current generation
        progeny = len(tables.individuals) + np.arange(N, dtype=np.int32)

        # NB this is *with* replacement, so 1 / N chance of selfing
        parents = rng.choice(population, (N, 2))
        tables.individuals.append_columns(
            flags=np.zeros(N, dtype=np.uint32),
            parents=parents.reshape(N * 2),
            parents_offset=np.arange(N + 1, dtype=np.uint32) * 2,
        )
        node_individual = np.zeros(2 * N, dtype=np.int32)
        node_individual[0::2] = progeny
        node_individual[1::2] = progeny
        flags_value = 0 if time > 0 else tskit.NODE_IS_SAMPLE
        tables.nodes.append_columns(
            flags=np.full(2 * N, flags_value, dtype=np.uint32),
            time=np.full(2 * N, time, dtype=np.float64),
            population=np.full(2 * N, 0, dtype=np.int32),
            individual=node_individual,
        )
        population = progeny

    return tables


#####
# All code below is scheduled for deletion:
# https://github.com/tskit-dev/msprime/issues/1856
#####

# NOTE This functionality is preliminary and undocumented. It
# *will change*.

# TODO the main thing we need to do here is get rid of the 1-based
# IDs and the "individual" column so that it maps directly into the
# individual table in tskit. We can integrate with PED files by
# putting in a null individual if necessary, but it's horribly
# confusing working with external IDs here.

# Another major issue here is that the sampling strategy is
# bound up with the representation of the pedigree itself. We should
# remove any idea of sampling from this class and instead make the
# samples argument to simulate specify *individual* IDs, ie.
# ts = msprime.sim_ancestry([0, 1], pedigree=pedigree)
# sets the individuals 0 and 1 to be the samples.


class Pedigree:
    """
    Class representing a pedigree for simulations.

    :param ndarray individual: A 1D integer array containing strictly positive unique IDs
        for each individual in the pedigree
    :param ndarray parents: A 2D integer array containing the indices, in the
        individual array, of an individual's parents
    :param ndarray time: A 1D float array containing the time of each
        individual, in generations in the past
    :param int ploidy: The ploidy of individuals in the pedigree. Currently
        only ploidy of 2 is supported
    """

    def __init__(self, individual, parents, time, is_sample=None, sex=None, ploidy=2):
        if ploidy != 2:
            raise NotImplementedError("Ploidy != 2 not currently supported")

        if sex is not None:
            raise NotImplementedError(
                "Assigning individual sexes not currently supported"
            )

        if parents.shape[1] != ploidy:
            raise ValueError(
                "Ploidy {} conflicts with number of parents {}".format(
                    ploidy, parents.shape[1]
                )
            )

        if np.min(individual) <= 0:
            raise ValueError("Individual IDs must be > 0")

        self.individual = individual.astype(np.int32)
        self.num_individuals = len(individual)
        self.parents = parents.astype(np.int32)
        self.time = time.astype(np.float64)
        self.sex = sex
        self.ploidy = int(ploidy)

        self.is_sample = None
        if is_sample is not None:
            self.is_sample = is_sample.astype(np.uint32)
            self.samples = np.array(is_sample)[np.where(is_sample == 1)]
            self.num_samples = len(self.samples)

    def set_samples(self, num_samples=None, sample_IDs=None, probands_only=True):
        if probands_only is not True:
            raise NotImplementedError("Only probands may currently be set as samples.")

        if num_samples is None and sample_IDs is None:
            raise ValueError("Must specify one of num_samples of sample_IDs")

        if num_samples is not None and sample_IDs is not None:
            raise ValueError("Cannot specify both samples and num_samples.")

        self.is_sample = np.zeros((self.num_individuals), dtype=np.uint32)

        all_indices = range(len(self.individual))
        proband_indices = set(all_indices).difference(self.parents.ravel())

        if num_samples is not None:
            self.num_samples = num_samples
            if self.num_samples > len(proband_indices):
                raise ValueError(
                    (
                        "Cannot specify more samples ({}) than there are "
                        "probands in the pedigree ({}) "
                    ).format(self.num_samples, len(proband_indices))
                )
            sample_indices = np.random.choice(
                list(proband_indices), size=self.num_samples, replace=False
            )

        elif sample_IDs is not None:
            self.num_samples = len(sample_IDs)

            indices = all_indices
            if probands_only:
                indices = proband_indices

            sample_set = set(sample_IDs)
            sample_indices = [i for i in indices if self.individual[i] in sample_set]

        if len(sample_indices) != self.num_samples:
            raise ValueError(
                "Sample size mismatch - duplicate sample IDs or sample ID not "
                "in pedigree"
            )

        self.is_sample[sample_indices] = 1

    def get_proband_indices(self):
        all_indices = range(len(self.individual))
        proband_indices = set(all_indices).difference(self.parents.ravel())

        return sorted(proband_indices)

    def get_ll_representation(self):
        """
        Returns the low-level representation of this Pedigree.
        """
        return {
            "individual": self.individual,
            "parents": self.parents,
            "time": self.time,
            "is_sample": self.is_sample,
        }

    # FIXME this shouldn't be a static method, but should be called automatically
    # if a time isn't specified.
    @staticmethod
    def get_times(individual, parent_IDs=None, parents=None, check=False):
        """
        For pedigrees without specified times, crudely assigns times to
        all individuals.
        """
        if parents is None and parent_IDs is None:
            raise ValueError("Must specify either parent IDs or parent indices")

        if parents is None:
            parents = Pedigree.parent_ID_to_index(individual, parent_IDs)

        time = np.zeros(len(individual))
        all_indices = range(len(individual))
        proband_indices = set(all_indices).difference(parents.ravel())
        climber_indices = proband_indices

        t = 0
        while len(climber_indices) > 0:
            next_climbers = []
            for c_idx in climber_indices:
                if time[c_idx] < t:
                    time[c_idx] = t

                next_parents = [p for p in parents[c_idx] if p >= 0]
                next_climbers.extend(next_parents)

            climber_indices = list(set(next_climbers))
            t += 1

        if check:
            Pedigree.check_times(individual, parents, time)

        return time

    @staticmethod
    def check_times(individual, parents, time):
        for i, ind in enumerate(individual):
            for parent_ix in parents[i]:
                if parent_ix >= 0:
                    t1 = time[i]
                    t2 = time[parent_ix]
                    if t1 >= t2:
                        raise ValueError(
                            "Ind {} has time >= than parent {}".format(
                                ind, individual[parent_ix]
                            )
                        )

    @staticmethod
    def parent_ID_to_index(individual, parent_IDs):
        n_inds, n_parents = parent_IDs.shape
        parents = np.zeros(parent_IDs.shape, dtype=int)
        ind_to_index_dict = dict(zip(individual, range(n_inds)))

        if 0 in ind_to_index_dict:
            raise ValueError(
                "Invalid ID: 0 reserved to denote individual" "not in the genealogy"
            )
        ind_to_index_dict[0] = -1

        for i in range(n_inds):
            for j in range(n_parents):
                parent_ID = parent_IDs[i, j]
                parents[i, j] = ind_to_index_dict[parent_ID]

        return parents

    @staticmethod
    def parent_index_to_ID(individual, parents):
        n_inds, n_parents = parents.shape
        parent_IDs = np.zeros(parents.shape, dtype=int)

        for i in range(n_inds):
            for j in range(n_parents):
                parent_ID = 0
                if parents[i, j] >= 0:
                    parent_ID = individual[parents[i, j]]
                    parent_IDs[i, j] = parent_ID

        return parent_IDs

    @staticmethod
    def default_format():
        cols = {
            "individual": 0,
            "parents": [1, 2],
            "time": 3,
            "is_sample": None,
            "sexes": None,
        }

        return cols

    @staticmethod
    def read_txt(pedfile, time_col=None, sex_col=None, **kwargs):
        """
        Creates a Pedigree instance from a text file.
        """
        cols = Pedigree.default_format()
        cols["time"] = time_col
        cols["sexes"] = sex_col

        if sex_col:
            raise NotImplementedError("Specifying sex of individuals not yet supported")

        usecols = []
        for c in cols.values():
            if isinstance(c, collections.abc.Iterable):
                usecols.extend(c)
            elif c is not None:
                usecols.append(c)
        usecols = sorted(usecols)

        data = np.genfromtxt(pedfile, skip_header=1, usecols=usecols, dtype=float)

        individual = data[:, cols["individual"]].astype(int)
        parent_IDs = data[:, cols["parents"]].astype(int)
        parents = Pedigree.parent_ID_to_index(individual, parent_IDs)

        if cols["time"] is not None:
            time = data[:, cols["time"]]
        else:
            time = Pedigree.get_times(individual, parents=parents)

        return Pedigree(individual, parents, time, **kwargs)

    @staticmethod
    def read_npy(pedarray_file, **kwargs):
        """
        Reads pedigree from numpy .npy file with columns:

            ind ID, father array index, mother array index, time

        where time is given in generations.
        """
        basename, ext = os.path.split(pedarray_file)
        pedarray = np.load(pedarray_file)

        cols = Pedigree.default_format()
        if "cols" in kwargs:
            cols = kwargs["cols"]

        individual = pedarray[:, cols["individual"]]
        parents = np.stack([pedarray[:, i] for i in cols["parents"]], axis=1)
        parents = parents.astype(int)
        time = pedarray[:, cols["time"]]

        P = Pedigree(individual, parents, time, **kwargs)

        return P

    def build_array(self):
        cols = Pedigree.default_format()

        col_nums = []
        for v in cols.values():
            if isinstance(v, collections.abc.Iterable):
                col_nums.extend(v)
            elif v is not None:
                col_nums.append(v)

        n_cols = max(col_nums) + 1

        if max(np.diff(col_nums)) > 1:
            raise ValueError(f"Non-sequential columns in pedigree format: {col_nums}")

        pedarray = np.zeros((self.num_individuals, n_cols))
        pedarray[:, cols["individual"]] = self.individual
        pedarray[:, cols["parents"]] = self.parents
        pedarray[:, cols["time"]] = self.time

        return pedarray

    def save_txt(self, fname):
        """
        Saves pedigree in text format with columns:

            ind ID, father array index, mother array index, time

        where time is given in generations.
        """
        pedarray = self.build_array()
        cols = self.default_format()
        cols_to_save = [
            cols["individual"],
            cols["parents"][0],
            cols["parents"][1],
            cols["time"],
        ]
        pedarray = pedarray[cols_to_save]
        parent_IDs = Pedigree.parent_index_to_ID(self.individual, self.parents)
        pedarray[:, cols["parents"]] = parent_IDs

        with open(fname, "w") as f:
            header = "ind\tfather\tmother\ttime\n"
            f.write(header)
            for row in pedarray:
                f.write("\t".join([str(x) for x in row]) + "\n")

    def save_npy(self, fname):
        """
        Saves pedigree in numpy .npy format with columns:

            ind ID, father array index, mother array index, time

        where time is given in generations.
        """
        pedarray = self.build_array()
        np.save(fname, pedarray)

    def asdict(self):
        """
        Returns a dict of arguments to recreate this pedigree
        """
        return {
            key: getattr(self, key)
            for key in inspect.signature(self.__init__).parameters.keys()
            if hasattr(self, key)
        }



#
# Copyright (C) 2015-2021 University of Oxford
#
# This file is part of msprime.
#
# msprime is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# msprime is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with msprime.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Core functions and classes used throughout msprime.
"""
from __future__ import annotations

import dataclasses
import html
import numbers
import os
import random
import textwrap
from typing import Any

import numpy as np

from msprime import _msprime

__version__ = "undefined"
try:
    from . import _version

    __version__ = _version.version
except ImportError:
    pass


# Make sure the GSL error handler is turned off so that we can be sure that
# we don't abort on errors. This can be reset by using the function
# _msprime.restore_gsl_error_handler(), which will set the error handler to
# the value that it had before this function was called.
_msprime.unset_gsl_error_handler()


# Some machinery here for generating default random seeds. We need a map
# indexed by process ID here because we cannot use a global variable
# to store the state across multiple processes. Copy-on-write semantics
# for child processes means that they inherit the state of the parent
# process, so if we just keep a global variable without indexing by
# PID, child processes will share the same random generator as the
# parent.

_seed_rng_map: dict[int, random.Random] = {}


def get_seed_rng() -> random.Random | None:
    return _seed_rng_map.get(os.getpid(), None)


def clear_seed_rng():
    _seed_rng_map.pop(os.getpid(), None)


def get_random_seed() -> int:
    global _seed_rng_map
    pid = os.getpid()
    if pid not in _seed_rng_map:
        # If we don't provide a seed to Random(), Python will seed either
        # from a system source of randomness (i.e., /dev/urandom) or the
        # current time if this is not available. Thus, our seed rng should
        # be unique, even across different processes.
        _seed_rng_map[pid] = random.Random()
    return _seed_rng_map[pid].randint(1, 2 ** 32 - 1)


def set_seed_rng_seed(seed: int):
    """
    Convenience method to let us make unseeded simulations deterministic
    when generating documentation examples.

    DO NOT USE THIS FUNCTION!!!
    """
    global _seed_rng_map
    pid = os.getpid()
    _seed_rng_map[pid] = random.Random(seed)


def isinteger(value: Any) -> bool:
    """
    Returns True if the specified value can be converted losslessly to an
    integer.
    """
    if isinstance(value, numbers.Number):
        # Mypy doesn't realise we've done an isinstance here.
        return int(value) == float(value)  # type: ignore
    return False


def _parse_flag(value: Any, *, default: bool) -> bool:
    """
    Parses a boolean flag, which can be either True, False, or None.
    If the input value is None, return the default. Otherwise,
    check that the input value is a bool.

    Note that we do *not* cast to a bool as this would accept
    truthy values like the empty list, etc. In this case None
    would be converted to False, potentially conflicting with
    the default value.
    """
    assert isinstance(default, bool)
    if value is None:
        return default
    if not isinstance(value, bool):
        raise TypeError("Boolean flag must be True, False, or None (the default value)")
    return value


@dataclasses.dataclass
class TableEntry:
    data: str
    extra: str | None = None

    def as_html(self, escape=True):
        ret = "<td"
        if self.extra is not None:
            wrapped = html.escape(textwrap.fill(self.extra, 80))
            ret += f" title='{wrapped}'"
        content = html.escape(self.data) if escape else self.data
        ret += f">{content}</td>"
        return ret

    def as_text(self):
        return self.data


def html_table(
    caption: str,
    column_titles: list[str],
    data: list[list[TableEntry | str]],
    no_escape: list[int] | None = None,
):
    """
    Returns a HTML table formatted with the specified data.
    """
    no_escape = [] if no_escape is None else no_escape
    header = "".join(f"<th>{col_title}</th>" for col_title in column_titles)
    rows = ""
    for row_data in data:
        assert len(column_titles) == len(row_data)
        row = ""
        for col_index, item in enumerate(row_data):
            if not isinstance(item, TableEntry):
                item = TableEntry(item)
            row += item.as_html(escape=col_index not in no_escape)
        rows += f"<tr>{row}</tr>"
    s = (
        "<div>"
        """<style scoped="">
            .tskit-table thead tr th:only-of-type {vertical-align: middle;}
            .tskit-table thead tr th {text-align: center;vertical-align: top;}
            .tskit-table tbody td {text-align: right;padding: 0.5em 0.5em;}
            .tskit-table tbody th {padding: 0.5em 0.5em;}
        </style>"""
        f"<b>{caption}</b>"
        '<table border="1" class="tskit-table">'
        "<thead>"
        "<tr>" + header + "</tr>"
        "</thead>"
        "<tbody>" + rows + "</tbody>"
        "</table>"
        "</div>"
    )
    return s


def _text_table_row(data, alignments, widths):
    num_lines = max(len(item) for item in data)
    for item in data:
        assert isinstance(item, list)
        item.extend([""] * (num_lines - len(item)))
        assert len(item) == num_lines
    s = ""
    for line in range(num_lines):
        out_line = "│"
        for value, align, width in zip(data, alignments, widths):
            out_line += f"{value[line]:{align}{width - 1}}│"
        out_line += "\n"
        s += out_line
    return s


def text_table(
    caption: str,
    column_titles: list[list[str]],
    column_alignments: list[str],
    data: list[list[list[str]]],
    internal_hlines=False,
):
    """
    Returns a text table formatted with the specified data. Column alignments
    should be values used in Python's string formatting mini-language.

    Each item in the table should be a *list* of strings, which are the lines
    of text to be shown in that table cell.
    """
    N = len(column_titles)
    assert len(column_alignments) == N
    widths = np.array([len(title) for title in column_titles], dtype=int)
    for row in data + [column_titles]:
        assert N == len(row)
        for j in range(N):
            widths[j] = max(widths[j], max((len(line) for line in row[j]), default=0))
    widths += 3

    hline = "─" * (sum(widths) - 1)
    internal_hline = "┈" * (sum(widths) - 1)
    out = f"{caption}\n"
    out += f"┌{hline}┐\n"
    out += f"{_text_table_row(column_titles, column_alignments, widths)}"
    out += f"├{hline}┤\n"
    for j, split_row in enumerate(data):
        out += f"{_text_table_row(split_row, column_alignments, widths)}"
        if internal_hlines and j < len(data) - 1:
            out += f"│{internal_hline}│\n"
    out += f"└{hline}┘\n"
    return out



#
# Copyright (C) 2015-2021 University of Oxford
#
# This file is part of msprime.
#
# msprime is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# msprime is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with msprime.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Command line interfaces to the msprime library.
"""
import argparse
import hashlib
import logging
import os
import random
import signal
import sys
import warnings
from typing import Dict
from typing import List

import demes
import tskit

import msprime
from . import ancestry
from . import mutations


logger = logging.getLogger(__name__)


def set_sigpipe_handler():
    if os.name == "posix":
        # Set signal handler for SIGPIPE to quietly kill the program.
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def setup_logging(args):
    log_level = "WARN"
    if args.verbosity > 0:
        log_level = "INFO"
    if args.verbosity > 1:
        log_level = "DEBUG"
    logging.basicConfig(
        level=log_level, format="%(asctime)s %(name)s %(levelname)s %(message)s"
    )


#######################################################
# mspms: the ms compatible interface
#######################################################

mscompat_description = (
    "mspms is an ms-compatible interface to the msprime library. "
    "It simulates the coalescent with recombination for a variety of "
    "demographic models and outputs the results in a text-based format. "
    "It supports a subset of the functionality available in ms and aims "
    "for full compatibility. "
    "WARNING: due to quirks in Python's argparse module, negative growth rates "
    "written in exponential form (e.g. `-eG 1.0 -1e-5`) are not recognised as an "
    "option argument. To work around this, specify the argument using quotes "
    "and a leading space, e.g. `-eG 1.0 ' -1e-5'`, or avoid scientific notation, "
    "e.g. `-eG 1.0 -0.00001`."
)
mscompat_recombination_help = (
    "Recombination at rate rho=4*N0*r where r is the rate of recombination "
    "between the ends of the region being simulated; num_loci is the number "
    "of sites between which recombination can occur"
)
mscompat_gene_conversion_help = (
    "Gene conversion at rate gamma where gamma depends on the defined "
    "recombination rate rho=4*N0*r. If rho > 0, gc_recomb_ratio defines the ratio "
    "g/r, where r is the probability per generation of crossing-over and g the "
    "corresponding gene conversion probability. Gene conversions are initiated at "
    "rate gamma=rho*gc_recomb_ratio = 4*N0*r*gc_recomb_ratio. If rho = 0 the gene "
    "conversion rate is given by gamma=gc_recomb_ratio=4*N0*c where c is the rate "
    "of gene conversion initiation between the ends of the simulated region of "
    "length num_loci. If the recombination rate is not specified, standard "
    "parameters are used, i.e. rho = 0 and num_loci = 1. The length of the gene "
    "conversion tracts is geometrically distributed with mean tract_length. "
    "The mean tract_length needs to be larger than or equal to 1 for discrete "
    "genomes and larger than 0 for continuous genomes."
)
mshotcompat_hotspot_help = (
    "Recombination hotspots defined according to the msHOT format. This is "
    "defined as a sequence: n (start stop scale)+ where n is the number of "
    "hotspots and each hotspot spans [start, stop) where the recombination "
    "rate is the background recombination rate times scale. Adjacent hotspots "
    "may stop and start at the same position but must otherwise be non-overlapping "
    "and specified in ascending order."
)

msprime_citation_text = """
If you use msprime in your work, please cite the following paper:
Jerome Kelleher, Alison M Etheridge and Gilean McVean (2016), "Efficient
Coalescent Simulation and Genealogical Analysis for Large Sample Sizes",
PLoS Comput Biol 12(5): e1004842. doi: 10.1371/journal.pcbi.1004842
"""


def positive_int(value):
    int_value = int(float(value))
    if int_value <= 0:
        msg = f"{value} in an invalid positive integer value"
        raise argparse.ArgumentTypeError(msg)
    return int_value


def add_sample_size_argument(parser):
    parser.add_argument(
        "sample_size", type=positive_int, help="The number of genomes in the sample"
    )


def add_random_seed_argument(parser):
    parser.add_argument(
        "--random-seed",
        "-s",
        type=int,
        default=None,
        help="The random seed; If not specified, one is chosen randomly.",
    )


def add_output_argument(parser):
    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("wb"),
        default=sys.stdout,
        help=(
            "Where to write the output tree sequence file. If omitted or '-', "
            "write to standard output"
        ),
    )


def generate_seeds():
    """
    Generate seeds to seed the RNG and output on the command line.
    """
    # Pull three numbers from the SystemRandom generator
    rng = random.SystemRandom()
    return [rng.randint(0, 2 ** 31) for _ in range(3)]


def get_single_seed(seeds):
    """
    Takes the specified command line seeds convert them to a single value
    that can be used to seed the python random number generator.
    """
    assert len(seeds) == 3
    m = hashlib.md5()
    for s in seeds:
        # Colon separate the values to ensure that we don't have
        # collisions in situations like 1:23:45, 12:3:45.
        m.update(f"{s}:".encode())
    # Now take the integer value of this modulo 2^32, as this is
    # the largest seed value we'll accept.
    return int(m.hexdigest(), 16) % (2 ** 32)


def hotspots_to_recomb_map(hotspots, background_rate, seq_length):
    """
    Translates the background recombination rate and recombination
    hotspots according to the msHOT cli spec to a msprime
    recombination map.

    hotspots is of the form: [n] + [start, stop, factor]+
    where n is the number of hotspots and each hotspot spans [start, stop)
    with a recombination rate of background_rate * factor. Intervals between
    hotspots have a recombination rate of background_rate.
    """
    assert len(hotspots) > 0
    n_hotspots = hotspots[0]
    assert len(hotspots[1:]) == 3 * n_hotspots
    positions = []
    rates = []
    if hotspots[1] != 0:
        positions.append(0)
        rates.append(background_rate)
    for i in range(1, len(hotspots) - 2, 3):
        [start, stop, factor] = hotspots[i : i + 3]
        # Beginning hotspot
        positions.append(start)
        rates.append(factor * background_rate)

        if i == len(hotspots) - 3 or stop != hotspots[i + 3]:
            # Ending hotspot, back to normal recombination rate
            positions.append(stop)
            if stop != seq_length:
                rates.append(background_rate)

    if positions[-1] != seq_length:
        positions.append(seq_length)

    return msprime.RateMap(position=positions, rate=rates)


class SimulationRunner:
    """
    Class to run msprime simulation and output the results.
    """

    def __init__(
        self,
        num_samples,
        demography=None,
        num_loci=1,
        recombination_rate=0,
        num_replicates=1,
        mutation_rate=0,
        print_trees=False,
        precision=3,
        ms_random_seeds=None,
        gene_conversion_rate=0,
        gene_conversion_tract_length=1,
        hotspots=None,
    ):
        self.num_loci = num_loci
        self.num_replicates = num_replicates
        self.mutation_rate = mutation_rate
        self.precision = precision
        self.print_trees = print_trees
        if hotspots is None:
            recomb_map = msprime.RateMap.uniform(num_loci, recombination_rate)
        else:
            recomb_map = hotspots_to_recomb_map(hotspots, recombination_rate, num_loci)

        if demography is None:
            # This is just used for testing so values don't really matter.
            demography = msprime.Demography.isolated_model([1])

        self.ms_random_seeds = ms_random_seeds
        if ms_random_seeds is None:
            self.ms_random_seeds = generate_seeds()
        # The ms command line requires three integers. We combine these into
        # a single seed.
        random_seed = get_single_seed(self.ms_random_seeds)

        # We need to get direct access to the simulator here because of the
        # "invisible" recombination breakpoints, so we can't run simulations
        # the usual way via sim_ancestry.
        self.simulator = ancestry._parse_sim_ancestry(
            samples=dict(enumerate(num_samples)),
            demography=demography,
            recombination_rate=recomb_map,
            gene_conversion_rate=gene_conversion_rate,
            gene_conversion_tract_length=gene_conversion_tract_length,
            ploidy=1,
            random_seed=random_seed,
        )

    def _print_trees(self, tree_sequence, output):
        """
        Print out the trees in ms-format from the specified tree sequence.
        When 'invisible' recombinations occur ms prints out copies of the
        same tree. Therefore, we must keep track of all breakpoints from the
        simulation and write out a tree for each one.
        """
        if self.num_loci == 1:
            tree = next(tree_sequence.trees())
            newick = tree.newick(precision=self.precision)
            print(newick, file=output)
        else:
            breakpoints = list(self.simulator.breakpoints) + [self.num_loci]
            j = 0
            for tree in tree_sequence.trees():
                newick = tree.newick(precision=self.precision)
                left, right = tree.interval
                while j < len(breakpoints) and breakpoints[j] <= right:
                    length = int(breakpoints[j] - left)
                    left = breakpoints[j]
                    j += 1
                    print(f"[{length}]", end="", file=output)
                    print(newick, file=output)

    def run(self, output):
        """
        Runs the simulations and writes the output to the specified
        file handle.
        """
        # The first line of ms's output is the command line.
        print(" ".join(sys.argv), file=output)
        print(" ".join(str(s) for s in self.ms_random_seeds), file=output)
        replicates = self.simulator.run_replicates(
            self.num_replicates,
            mutation_rate=self.mutation_rate,
        )
        for ts in replicates:
            print(file=output)
            print("//", file=output)
            if self.print_trees:
                self._print_trees(ts, output)
            if self.mutation_rate > 0:
                assert ts.num_sites == ts.num_mutations
                s = ts.num_sites
                print("segsites:", s, file=output)
                if s != 0:
                    print("positions: ", end="", file=output)
                    for site in ts.sites():
                        x = site.position / self.num_loci
                        print(
                            "{0:.{1}f}".format(x, self.precision),
                            end=" ",
                            file=output,
                        )
                    print(file=output)
                    for h in ts.haplotypes():
                        print(h, file=output)
                else:
                    print(file=output)


def convert_int(value, parser):
    """
    Converts the specified value to an integer if possible. If
    conversion fails, exit by calling parser.error.
    """
    try:
        return int(value)
    except ValueError:
        parser.error(f"invalid int value '{value}'")


def convert_float(value, parser):
    """
    Converts the specified value to a float if possible. If
    conversion fails, exit by calling parser.error.
    """
    try:
        return float(value)
    except ValueError:
        parser.error(f"invalid float value '{value}'")


def convert_population_id(parser, population_id, num_populations):
    """
    Checks the specified population ID makes sense and returns
    it as an integer, and converted into msprime's internal population
    ID scheme (i.e., zero based).
    """
    pid = int(population_id)
    if pid != population_id:
        msg = "Bad population ID '{}': must be an integer"
        parser.error(msg.format(population_id))
    if pid < 1 or pid > num_populations:
        msg = "Bad population ID '{}': must be 1 to num_populations"
        parser.error(msg.format(pid))
    return pid - 1


def check_migration_rate(parser, rate):
    """
    Checks that the specified migration rate makes sense.
    """
    if rate < 0:
        parser.error("Migration rates must be non-negative.")


def check_event_time(parser, time):
    """
    Checks that the specified time for an event makes sense
    """
    if time < 0:
        parser.error("Event times must be non-negative.")


def convert_migration_matrix(parser, input_matrix, num_populations):
    """
    Converts the specified migration matrix into the internal format.
    """
    if len(input_matrix) != num_populations ** 2:
        parser.error("Must be num_populations^2 migration matrix entries")
    migration_matrix = [
        [0 for j in range(num_populations)] for k in range(num_populations)
    ]
    for j in range(num_populations):
        for k in range(num_populations):
            if j != k:
                rate = convert_float(input_matrix[j * num_populations + k], parser)
                check_migration_rate(parser, rate)
                migration_matrix[j][k] = rate
    return migration_matrix


def raise_admixture_incompatability_error(parser, other_option):
    """
    Because of the state dependency within ms, it is messy for us to
    support options that affect all populations in conjunction with
    population splits. For now, raise an error.
    """
    parser.error(
        "Cannot currently use the -es and {} options together. "
        "Please open an issue on GitHub if this functionality is "
        "important to you.".format(other_option)
    )


def create_simulation_runner(parser, arg_list):
    """
    Parses the arguments and returns a SimulationRunner instance.
    """
    args = parser.parse_args(arg_list)
    if args.mutation_rate == 0 and not args.trees:
        parser.error("Need to specify at least one of --theta or --trees")
    num_loci = int(args.recombination[1])
    if args.recombination[1] != num_loci:
        parser.error("Number of loci must be integer value")
    if args.recombination[0] != 0.0 and num_loci < 2:
        parser.error("Number of loci must > 1")
    r = 0.0
    # We don't scale recombination or mutation rates by the size
    # of the region.
    if num_loci > 1:
        r = args.recombination[0] / (num_loci - 1)
    mu = args.mutation_rate / num_loci

    # ms uses a ratio to define the GC rate, but if the recombination rate
    # is zero we define the gc rate directly.
    gc_param, gc_tract_length = args.gene_conversion
    gc_rate = 0
    if r == 0.0:
        if num_loci > 1:
            gc_rate = gc_param / (num_loci - 1)
    else:
        gc_rate = r * gc_param

    demography = msprime.Demography.isolated_model([1])
    # Check the structure format.
    symmetric_migration_rate = 0.0
    num_populations = 1
    migration_matrix = [[0.0]]
    num_samples = [args.sample_size]
    if args.structure is not None:
        num_populations = convert_int(args.structure[0], parser)
        # We must have at least num_population sample_configurations
        if len(args.structure) < num_populations + 1:
            parser.error("Must have num_populations sample sizes")
        demography = msprime.Demography.isolated_model([1] * num_populations)
        num_samples = [0] * num_populations
        for j in range(num_populations):
            num_samples[j] = convert_int(args.structure[j + 1], parser)
        if sum(num_samples) != args.sample_size:
            parser.error("Population sample sizes must sum to sample_size")
        # We optionally have the overall migration_rate here
        if len(args.structure) == num_populations + 2:
            symmetric_migration_rate = convert_float(
                args.structure[num_populations + 1], parser
            )
            check_migration_rate(parser, symmetric_migration_rate)
        elif len(args.structure) > num_populations + 2:
            parser.error("Too many arguments to --structure/-I")
        if num_populations > 1:
            migration_matrix = [
                [
                    symmetric_migration_rate / (num_populations - 1) * int(j != k)
                    for j in range(num_populations)
                ]
                for k in range(num_populations)
            ]
    else:
        if len(args.migration_matrix_entry) > 0:
            parser.error(
                "Cannot specify migration matrix entries without "
                "first providing a -I option"
            )
        if args.migration_matrix is not None:
            parser.error(
                "Cannot specify a migration matrix without "
                "first providing a -I option"
            )
    if args.migration_matrix is not None:
        migration_matrix = convert_migration_matrix(
            parser, args.migration_matrix, num_populations
        )
    for matrix_entry in args.migration_matrix_entry:
        pop_i = convert_population_id(parser, matrix_entry[0], num_populations)
        pop_j = convert_population_id(parser, matrix_entry[1], num_populations)
        rate = matrix_entry[2]
        if pop_i == pop_j:
            parser.error("Cannot set diagonal elements in migration matrix")
        check_migration_rate(parser, rate)
        migration_matrix[pop_i][pop_j] = rate

    # Set the initial demography
    if args.growth_rate is not None:
        for population in demography.populations:
            population.growth_rate = args.growth_rate
    for population_id, growth_rate in args.population_growth_rate:
        pid = convert_population_id(parser, population_id, num_populations)
        demography.populations[pid].growth_rate = growth_rate
    for population_id, size in args.population_size:
        pid = convert_population_id(parser, population_id, num_populations)
        demography.populations[pid].initial_size = size

    demographic_events = []
    # First we look at population split events. We do this differently
    # to ms, as msprime requires a fixed number of population. Therefore,
    # modify the number of populations to take into account populations
    # splits. This is a messy hack, and will probably need to be changed.
    for index, (t, population_id, proportion) in args.admixture:
        check_event_time(parser, t)
        pid = convert_population_id(parser, population_id, num_populations)
        if proportion < 0 or proportion > 1:
            parser.error("Proportion value must be 0 <= p <= 1.")
        # In ms, the probability of staying in source is p and the probability
        # of moving to the new population is 1 - p.
        event = (index, msprime.MassMigration(t, pid, num_populations, 1 - proportion))
        demographic_events.append(event)

        num_populations += 1
        # We add another element to each row in the migration matrix
        # along with an other row. All new entries are zero.
        for row in migration_matrix:
            row.append(0)
        migration_matrix.append([0 for j in range(num_populations)])
        demography.add_population(initial_size=1)
        num_samples.append(0)

    # Add the demographic events
    for index, (t, alpha) in args.growth_rate_change:
        if len(args.admixture) != 0:
            raise_admixture_incompatability_error(parser, "-eG")
        check_event_time(parser, t)
        demographic_events.append(
            (index, msprime.PopulationParametersChange(time=t, growth_rate=alpha))
        )
    for index, (t, population_id, alpha) in args.population_growth_rate_change:
        pid = convert_population_id(parser, population_id, num_populations)
        check_event_time(parser, t)
        demographic_events.append(
            (
                index,
                msprime.PopulationParametersChange(
                    time=t, growth_rate=alpha, population_id=pid
                ),
            )
        )
    for index, (t, x) in args.size_change:
        if len(args.admixture) != 0:
            raise_admixture_incompatability_error(parser, "-eN")
        check_event_time(parser, t)
        demographic_events.append(
            (
                index,
                msprime.PopulationParametersChange(
                    time=t, initial_size=x, growth_rate=0
                ),
            )
        )
    for index, (t, population_id, x) in args.population_size_change:
        check_event_time(parser, t)
        pid = convert_population_id(parser, population_id, num_populations)
        demographic_events.append(
            (
                index,
                msprime.PopulationParametersChange(
                    time=t, initial_size=x, growth_rate=0, population_id=pid
                ),
            )
        )
    for index, (t, pop_i, pop_j) in args.population_split:
        check_event_time(parser, t)
        pop_i = convert_population_id(parser, pop_i, num_populations)
        pop_j = convert_population_id(parser, pop_j, num_populations)
        demographic_events.append((index, msprime.MassMigration(t, pop_i, pop_j, 1.0)))
        # Migration rates from subpopulation i (M[k, i], k != i) are set to zero.
        for k in range(num_populations):
            if k != pop_i:
                event = msprime.MigrationRateChange(t, 0.0, matrix_index=(k, pop_i))
                demographic_events.append((index, event))

    # Demographic events that affect the migration matrix
    if num_populations == 1:
        condition = (
            len(args.migration_rate_change) > 0
            or len(args.migration_matrix_entry_change) > 0
            or len(args.migration_matrix_change) > 0
        )
        if condition:
            parser.error("Cannot change migration rates for 1 population")
    for index, (t, x) in args.migration_rate_change:
        if len(args.admixture) != 0:
            raise_admixture_incompatability_error(parser, "-eM")
        check_migration_rate(parser, x)
        check_event_time(parser, t)
        event = msprime.MigrationRateChange(t, x / (num_populations - 1))
        demographic_events.append((index, event))
    for index, event in args.migration_matrix_entry_change:
        t = event[0]
        check_event_time(parser, t)
        pop_i = convert_population_id(parser, event[1], num_populations)
        pop_j = convert_population_id(parser, event[2], num_populations)
        if pop_i == pop_j:
            parser.error("Cannot set diagonal elements in migration matrix")
        rate = event[3]
        check_migration_rate(parser, rate)
        msp_event = msprime.MigrationRateChange(t, rate, matrix_index=(pop_i, pop_j))
        demographic_events.append((index, msp_event))
    for index, event in args.migration_matrix_change:
        if len(event) < 3:
            parser.error("Need at least three arguments to -ma")
        if len(args.admixture) != 0:
            raise_admixture_incompatability_error(parser, "-ema")
        t = convert_float(event[0], parser)
        check_event_time(parser, t)
        if convert_int(event[1], parser) != num_populations:
            parser.error("num_populations must be equal for new migration matrix")
        matrix = convert_migration_matrix(parser, event[2:], num_populations)
        for j in range(num_populations):
            for k in range(num_populations):
                if j != k:
                    msp_event = msprime.MigrationRateChange(
                        t, matrix[j][k], matrix_index=(j, k)
                    )
                    demographic_events.append((index, msp_event))

    demographic_events.sort(key=lambda x: (x[0], x[1].time))
    time_sorted = sorted(demographic_events, key=lambda x: x[1].time)
    if demographic_events != time_sorted:
        parser.error("Demographic events must be supplied in non-decreasing time order")
    for _, event in demographic_events:
        demography.add_event(event)
    demography.migration_matrix[:] = migration_matrix

    # Adjust the population sizes so that the timescales agree. In principle
    # we could correct this with a ploidy value=0.5, but what we have here
    # seems less awful.
    for msp_event in demography.events:
        if isinstance(msp_event, msprime.PopulationParametersChange):
            if msp_event.initial_size is not None:
                msp_event.initial_size /= 2
    for pop in demography.populations:
        pop.initial_size /= 2

    runner = SimulationRunner(
        num_samples,
        demography,
        num_loci=num_loci,
        num_replicates=args.num_replicates,
        recombination_rate=r,
        mutation_rate=mu,
        gene_conversion_rate=gc_rate,
        gene_conversion_tract_length=gc_tract_length,
        precision=args.precision,
        print_trees=args.trees,
        ms_random_seeds=args.random_seeds,
        hotspots=args.hotspots,
    )
    return runner


class IndexedAction(argparse._AppendAction):
    """
    Argparse action class that allows us to find the overall ordering
    across several different options. We use this for the demographic
    events, as the order in which the events are applied matters for
    ms compatibility.
    """

    index = 0

    def __call__(self, parser, namespace, values, option_string=None):
        super().__call__(
            parser, namespace, (IndexedAction.index, values), option_string
        )
        IndexedAction.index += 1


def convert_arg_line_to_args(arg_line):
    # from the docs on argparse.ArgumentParser.convert_arg_line_to_args
    return arg_line.split()


def make_load_file_action(next_parser):
    """
    Argparse action class to allow passing a filename containing arguments
    on the command line (for super-long argument files).
    From
        http://stackoverflow.com/q/27433316
    and
        http://stackoverflow.com/q/40060571
    """

    class LoadFromFile(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            try:
                with open(values) as f:
                    # note parses with 'next_parser' *not* with parser that is
                    # passed in
                    next_parser.parse_args(f.read().split(), namespace)
            except OSError as ioe:
                parser.error(ioe)

    return LoadFromFile


def get_mspms_parser(error_handler=None):
    # Ensure that the IndexedAction counter is set to zero. This is useful
    # for testing where we'll be creating lots of these parsers.
    IndexedAction.index = 0
    # to allow `-f` options we'll need a parser that can do all the
    # arguments except the positional (nonoptional) arguments.  We'll
    # create this one first, then at the end make the parser that
    # includes the positional arguments.
    parser = argparse.ArgumentParser(
        description=mscompat_description, epilog=msprime_citation_text
    )
    parser.convert_arg_line_to_args = convert_arg_line_to_args

    group = parser.add_argument_group("Behaviour")
    group.add_argument(
        "--mutation-rate",
        "-t",
        type=float,
        metavar="theta",
        help="Mutation rate theta=4*N0*mu",
        default=0,
    )
    group.add_argument(
        "--trees", "-T", action="store_true", help="Print out trees in Newick format"
    )
    group.add_argument(
        "--recombination",
        "-r",
        type=float,
        nargs=2,
        default=(0, 1),
        metavar=("rho", "num_loci"),
        help=mscompat_recombination_help,
    )
    group.add_argument(
        "--gene-conversion",
        "-c",
        type=float,
        nargs=2,
        default=(0, 1),
        metavar=("gc_recomb_ratio", "tract_length"),
        help=mscompat_gene_conversion_help,
    )
    group.add_argument(
        "--hotspots",
        "-v",
        type=float,
        nargs="+",
        default=None,
        help=mshotcompat_hotspot_help,
    )

    group = parser.add_argument_group("Structure and migration")
    group.add_argument(
        "--structure",
        "-I",
        nargs="+",
        metavar="value",
        help=(
            "Sample from populations with the specified deme structure. "
            "The arguments are of the form 'num_populations "
            "n1 n2 ... [4N0m]', specifying the number of populations, "
            "the sample configuration, and optionally, the migration "
            "rate for a symmetric island model"
        ),
    )
    group.add_argument(
        "--migration-matrix-entry",
        "-m",
        action="append",
        metavar=("i", "j", "rate"),
        nargs=3,
        type=float,
        default=[],
        help=(
            "Sets an entry M[i, j] in the migration matrix to the "
            "specified rate. i and j are (1-indexed) population "
            "IDs. Multiple options can be specified."
        ),
    )
    group.add_argument(
        "--migration-matrix",
        "-ma",
        nargs="+",
        default=None,
        metavar="entry",
        help=(
            "Sets the migration matrix to the specified value. The "
            "entries are in the order M[1,1], M[1, 2], ..., M[2, 1],"
            "M[2, 2], ..., M[N, N], where N is the number of populations."
        ),
    )
    group.add_argument(
        "--migration-rate-change",
        "-eM",
        nargs=2,
        action=IndexedAction,
        type=float,
        default=[],
        metavar=("t", "x"),
        help=(
            "Set the symmetric island model migration rate to "
            "x / (npop - 1) at time t"
        ),
    )
    group.add_argument(
        "--migration-matrix-entry-change",
        "-em",
        action=IndexedAction,
        metavar=("time", "i", "j", "rate"),
        nargs=4,
        type=float,
        default=[],
        help=(
            "Sets an entry M[i, j] in the migration matrix to the "
            "specified rate at the specified time. i and j are "
            "(1-indexed) population IDs."
        ),
    )
    group.add_argument(
        "--migration-matrix-change",
        "-ema",
        nargs="+",
        default=[],
        action=IndexedAction,
        metavar="entry",
        help=(
            "Sets the migration matrix to the specified value at time t."
            "The entries are in the order M[1,1], M[1, 2], ..., M[2, 1],"
            "M[2, 2], ..., M[N, N], where N is the number of populations."
        ),
    )

    group = parser.add_argument_group("Demography")
    group.add_argument(
        "--growth-rate",
        "-G",
        metavar="alpha",
        type=float,
        help=(
            "Set the growth rate to alpha for all populations. "
            "See warning above about negative growth rates."
        ),
    )
    group.add_argument(
        "--population-growth-rate",
        "-g",
        action="append",
        default=[],
        nargs=2,
        metavar=("population_id", "alpha"),
        type=float,
        help=(
            "Set the growth rate to alpha for a specific population. "
            "See warning above about negative growth rates."
        ),
    )
    group.add_argument(
        "--population-size",
        "-n",
        action="append",
        default=[],
        nargs=2,
        metavar=("population_id", "size"),
        type=float,
        help="Set the size of a specific population to size*N0.",
    )

    group.add_argument(
        "--growth-rate-change",
        "-eG",
        nargs=2,
        action=IndexedAction,
        type=float,
        default=[],
        metavar=("t", "alpha"),
        help=(
            "Set the growth rate for all populations to alpha at time t. "
            "See warning above about negative growth rates."
        ),
    )
    group.add_argument(
        "--population-growth-rate-change",
        "-eg",
        nargs=3,
        action=IndexedAction,
        type=float,
        default=[],
        metavar=("t", "population_id", "alpha"),
        help=(
            "Set the growth rate for a specific population to alpha at time t. "
            "See warning above about negative growth rates."
        ),
    )
    group.add_argument(
        "--size-change",
        "-eN",
        nargs=2,
        action=IndexedAction,
        type=float,
        default=[],
        metavar=("t", "x"),
        help="Set the population size for all populations to x * N0 at time t",
    )
    group.add_argument(
        "--population-size-change",
        "-en",
        nargs=3,
        action=IndexedAction,
        type=float,
        default=[],
        metavar=("t", "population_id", "x"),
        help=(
            "Set the population size for a specific population to " "x * N0 at time t"
        ),
    )
    group.add_argument(
        "--population-split",
        "-ej",
        nargs=3,
        action=IndexedAction,
        type=float,
        default=[],
        metavar=("t", "i", "j"),
        help=(
            "Move all lineages in population i to j at time t. "
            "Forwards in time, this corresponds to a population split "
            "in which lineages in j split into i. All migration "
            "rates for population i are set to zero."
        ),
    )
    group.add_argument(
        "--admixture",
        "-es",
        nargs=3,
        action=IndexedAction,
        type=float,
        default=[],
        metavar=("t", "population_id", "proportion"),
        help=(
            "Split the specified population into a new population, such "
            "that the specified proportion of lineages remains in "
            "the population population_id. Forwards in time this "
            "corresponds to an admixture event. The new population has ID "
            "num_populations + 1. Migration rates to and from the new "
            "population are set to 0, and growth rate is 0 and the "
            "population size for the new population is N0."
        ),
    )

    group = parser.add_argument_group("Miscellaneous")
    group.add_argument(
        "--random-seeds",
        "-seeds",
        nargs=3,
        type=positive_int,
        metavar=("x1", "x2", "x3"),
        help="Random seeds (must be three integers)",
    )
    group.add_argument(
        "--precision",
        "-p",
        type=positive_int,
        default=3,
        help="Number of values after decimal place to print",
    )

    # now for the parser that gets called first
    init_parser = argparse.ArgumentParser(
        description=mscompat_description,
        epilog=msprime_citation_text,
        add_help=False,
        parents=[parser],
    )
    init_parser.convert_arg_line_to_args = convert_arg_line_to_args

    add_sample_size_argument(init_parser)
    init_parser.add_argument(
        "num_replicates", type=positive_int, help="Number of independent replicates"
    )
    init_parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {msprime.__version__}"
    )
    init_parser.add_argument(
        "-f",
        "--filename",
        action=make_load_file_action(parser),
        help="Insert commands from a file at this point in the command line.",
    )

    # Set the optional error handler (used for testing)
    if error_handler is not None:
        parser.error = error_handler
        init_parser.error = error_handler

    return init_parser


def get_mspms_runner(arg_list):
    parser = get_mspms_parser()
    return create_simulation_runner(parser, arg_list)


def mspms_main(arg_list=None):
    set_sigpipe_handler()
    sr = get_mspms_runner(arg_list)
    sr.run(sys.stdout)


#######################################################
# msp: the command line interface for msprime
#######################################################


def run_simulate(args):
    warnings.warn(
        "The simulate command is deprecated but will continue to work "
        "indefinitely. To simulate ancestries, please use the 'ancestry' "
        "command. To place mutations on a tree sequence, please use "
        "the 'mutations' command."
    )

    if args.compress:
        warnings.warn(
            "The --compress option is no longer supported and does nothing. "
            "Please use the tszip utility to compress the output instead."
        )
    tree_sequence = msprime.simulate(
        sample_size=int(args.sample_size),
        Ne=args.effective_population_size,
        length=args.length,
        recombination_rate=args.recombination_rate,
        mutation_rate=args.mutation_rate,
        random_seed=args.random_seed,
    )
    tree_sequence.dump(args.tree_sequence)


def run_mutations(args):
    input_ts = tskit.load(args.input)
    output_ts = msprime.sim_mutations(
        tree_sequence=input_ts,
        rate=args.mutation_rate,
        random_seed=args.random_seed,
        keep=True,
        start_time=args.start_time,
        end_time=args.end_time,
        discrete_genome=True,
        model=args.model,
    )
    output_ts.dump(args.output)


def parse_sample_size(s: str) -> int:
    """
    Support sample sizes specified like 1e3, as well as integer literals
    like 1_000.
    """
    float_val = float(s)
    int_val = int(float_val)
    if int_val != float_val:
        raise ValueError(f"Bad sample size: {s}; must be an integer")
    return int_val


def parse_sample_spec(sample_spec: List[str]) -> Dict[str, int]:
    """
    Parse a list of <population_name>:<sample_size> values into a dictionary.
    """
    samples = {}
    for pop_spec in sample_spec:
        # We only accept ID:num_samples values
        tokens = pop_spec.split(":")
        if len(tokens) != 2 or any(len(tok) == 0 for tok in tokens):
            raise ValueError(
                "Sample specifications must be in the form <population_name:num_samples>"
            )
        name = tokens[0]
        if name in samples:
            raise ValueError(
                f"Population '{name}' referred to more than once in sample "
                "specification"
            )
        samples[name] = parse_sample_size(tokens[1])
    return samples


def run_ancestry(args):
    setup_logging(args)
    if args.demography is not None:
        graph = demes.load(args.demography)
        logger.debug(f"Loaded demes graph from {args.demography}")
        demography = msprime.Demography.from_demes(graph)
        samples = parse_sample_spec(args.samples)
    else:
        if len(args.samples) != 1:
            raise ValueError("Samples must be a single integer")
        demography = None
        samples = parse_sample_size(args.samples[0])
    tree_sequence = msprime.sim_ancestry(
        samples=samples,
        demography=demography,
        population_size=args.population_size,
        sequence_length=args.length,
        ploidy=args.ploidy,
        recombination_rate=args.recombination_rate,
        random_seed=args.random_seed,
    )
    tree_sequence.dump(args.output)


def get_msp_parser():
    top_parser = argparse.ArgumentParser(
        description="Command line interface for msprime.", epilog=msprime_citation_text
    )
    top_parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {msprime.__version__}"
    )
    subparsers = top_parser.add_subparsers(dest="subcommand")
    subparsers.required = True

    add_ancestry_subcommand(subparsers)
    add_mutate_subcommand(subparsers)
    add_simulate_subcommand(subparsers)

    return top_parser


def add_mutate_subcommand(subparsers):
    parser = subparsers.add_parser(
        # see note above on aliases and sphinx-argparse
        # aliases=["mut"],
        "mutations",
        help="Simulate mutations on a tree sequence.",
    )
    parser.add_argument(
        "mutation_rate",
        type=float,
        help="The mutation rate per base",
    )
    parser.add_argument(
        "input",
        nargs="?",
        type=argparse.FileType("rb"),
        default=sys.stdin,
        help=("The input tree sequence. If omitted or '-' read from standard input"),
    )
    add_output_argument(parser)
    add_random_seed_argument(parser)
    parser.add_argument(
        "--start-time",
        type=float,
        default=None,
        help="The minimum time ago at which a mutation can occur",
    )
    parser.add_argument(
        "--end-time",
        type=float,
        default=None,
        help="The maximum time ago at which a mutation can occur",
    )
    parser.add_argument(
        "--model",
        "-m",
        type=str,
        default="jc69",
        choices=sorted(mutations.MODEL_MAP.keys()),
        help="The mutation model to use (default=JC69)",
    )
    parser.set_defaults(runner=run_mutations)


def add_simulate_subcommand(subparsers) -> None:
    parser = subparsers.add_parser(
        "simulate",
        help=(
            "Run a simulation. DEPRECATED - please use `msp ancestry` "
            "and `msp mutations` instead"
        ),
    )
    add_sample_size_argument(parser)
    parser.add_argument("tree_sequence", help="The output tree sequence file")
    parser.add_argument(
        "--length",
        "-L",
        type=float,
        default=1,
        help="The length of the simulated region in base pairs",
    )
    parser.add_argument(
        "--recombination-rate",
        "-r",
        type=float,
        default=0,
        help="The recombination rate per base per generation",
    )
    parser.add_argument(
        "--mutation-rate",
        "-u",
        type=float,
        default=0,
        help="The mutation rate per base per generation",
    )
    parser.add_argument(
        "--effective-population-size",
        "-N",
        type=float,
        default=1,
        help="The diploid effective population size Ne",
    )
    add_random_seed_argument(parser)
    parser.add_argument(
        "--compress",
        "-z",
        action="store_true",
        help="Deprecated option with no effect; please use the tszip utility instead.",
    )
    parser.set_defaults(runner=run_simulate)


def add_ancestry_subcommand(subparsers) -> None:
    parser = subparsers.add_parser(
        "ancestry",
        # We'd like to use aliases there but sphinx-argparse doesn't support it
        # in the released version. Looks like the project has been abandoned by
        # its original maintainers, so we might want to stop using it.
        # aliases=["anc"],
        help="Simulate an ancestral history and output as a tskit tree sequence.",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        action="count",
        default=0,
        help="Increase the verbosity. Use -v for INFO output and -vv for DEBUG",
    )
    parser.add_argument(
        "samples",
        nargs="+",
        help=(
            "The sample specification. If a demography is not specified using the "
            "-d option, this must be a single integer denoting the number of "
            "k-ploid individuals (see the --ploidy argument) to sample. "
            "If a demography is specified, the samples must be provided as "
            "<population identifier>:<num samples> pairs. Samples from multiple "
            "populations can be specified; for example, if we have a demography "
            "with two populations named A and B the sample specification 'A:5 B:6' will "
            "sample 5 individuals from A and 6 from B. Samples are always taken at "
            "the corresponding population's default sampling time."
        ),
    )
    add_output_argument(parser)
    add_random_seed_argument(parser)
    parser.add_argument(
        "--length",
        "-L",
        type=float,
        default=1,
        help="The length of the genome sequence to simulate",
    )
    parser.add_argument(
        "--recombination-rate",
        "-r",
        type=float,
        default=0,
        help="The recombination rate per base per generation",
    )
    parser.add_argument(
        "--ploidy",
        "-k",
        type=int,
        default=2,
        help="The number of monoploid genomes per sample individual",
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--population-size",
        "-N",
        type=float,
        default=None,
        help="The number of individuals in the population",
    )
    group.add_argument(
        "--demography",
        "-d",
        type=str,
        default=None,
        help="The path to a Demes YAML file describing the demographic model.",
    )
    parser.set_defaults(runner=run_ancestry)


def msp_main(arg_list=None):
    set_sigpipe_handler()
    parser = get_msp_parser()
    args = parser.parse_args(arg_list)
    args.runner(args)
